#ifndef IMDI_GEN_H
#define IMDI_GEN_H

/* Integer Multi-Dimensional Interpolation */

/*
 * Copyright 2000 - 2006 Graeme W. Gill
 * All rights reserved.
 *
 * This material is licenced under the GNU GENERAL PUBLIC LICENCE :-
 * see the Licence.txt file for licencing details.
 */

/*
 * Pixel code generation definitions.
 * 
 * This defines a particular combination of pixel layout,
 * number of channels, and other runtime requestable features.
 * This is used by the code generation to setup anticipated
 * support for these requests, used by the runtime to request
 * the features, and by the runtime to identify the correct
 * generated module to use.
 *
 */


/* -------------------------------------------------- */
/* High level kernel generation description */

/* Processing options */
typedef enum {
	opts_none      = 0x00,	/* Forward direction, no stride */
	opts_fwd       = 0x01,	/* Forwards direction (default is gen fwd/run don't care) */
	opts_bwd       = 0x02,	/* Backwards direction (default is gen fwd/run don't care) */
	opts_istride   = 0x04,	/* Stride on input */
	opts_ostride   = 0x08,	/* Stride on output */
	opts_end       = 0x80	/* End marker */
} imdi_options;


/* This is a high level macro desciption of the pixel layout. */
/* It can be expanded by adding a new enumeration, and then */
/* implementing the code in imdi_gen to translate the enumeration */
/* into the exact pixlayout structure details. */

typedef enum {
	invalid_rep = 0x00,
	pixint8     = 0x01,		/* 8 Bits per value, pixel interleaved, no padding */
	planeint8   = 0x02,		/* 8 bits per value, plane interleaved */
	pixint16    = 0x03,		/* 16 Bits per value, pixel interleaved, no padding */
	planeint16  = 0x04		/* 16 bits per value, plane interleaved */
} imdi_pixrep;

/* The internal processing precision */
typedef enum {
	prec_min   = 0,		/* Minimum of input and output precision */
	prec_max   = 1,		/* Maximum of input and output precision */
	prec_in    = 2,		/* Same as input */
	prec_out   = 3,		/* Same as output */
	prec_p8    = 4,		/* 8 Bits precision */
	prec_p16   = 5		/* 16 Bits precision */
} imdi_iprec;

/* This sructure allows a series of related kernels to be generated */
#define MX_GDCS 20

typedef struct {
	int idcombs[MX_GDCS];			/* Input dimension combinations (0 at end) */
	int itres[MX_GDCS];				/* Interpolation table resolutions for each input dimension */

	int stres[MX_GDCS];				/* Simplex table resolutions for each input dimension */
	int odcombs[MX_GDCS];			/* Output dimensions combinations (0 at end) */

	imdi_pixrep incombs[MX_GDCS];	/* Input pixel representation combinations */
	imdi_iprec  iprecs[MX_GDCS];	/* Internal precision for each input representation */
	imdi_pixrep outcombs[MX_GDCS];	/* Output pixel representation for each input representation */

	imdi_options optcombs[MX_GDCS];	/* Direction and stride combinations */
} gendesc;

/* -------------------------------------------------- */
/* Detailed level of generation specification */

/* Pixel layout: */
/* Each channel value is assumed to be read with a 		*/
/* native machine single read of size bpch, 			*/
/* and bov and bpv being bit indexes into that value.	*/
/*                                        				*/
/* If pint == 0, then each read will be of size bpch[], and will */
/* have its own pointer, will be incremented by chi[].	*/
/* If pint != 0, then the reads will be size bpch[0], from successive */
/* locations, chi[] apart. 								*/
/*                                        				*/
/* If packed == 0, then separate reads are needed for each input */
/* channel.												*/
/* If packed != 0, then the channel values will be extracted */
/* from a single read of size bpch[0]					*/
/*                                        				*/
/* Note that at all times the bit offset and size values */
/* will be obeyed for each input value. 				*/

/* NOTE :- if you change this, you need to change the code in cgen.c */
/* labeled !genspec and tabspec delta code! */
typedef struct {
	int bpch[IXDIDO];	/* Bits per channel read (must be divisible by 8) */
	int chi[IXDIDO];	/* channel increment in multiples of bpch[] (0 == dimensionality) */
	int bov[IXDIDO];	/* Bit offset to value within channel */
	int	bpv[IXDIDO];	/* Bits per value within channel */
	int pint;			/* Flag - nonz if pixel interleaved (ie. reads from successice locations) */
	int packed;			/* Flag - nonz if all channels are packed into a single read */
} pixlayout;

/* Structure that specifies the configuration of a generated interpolation kernel. */
/* NOTE :- if you change this, you need to change the code in cgen.c */
/* labeled !genspec and tabspec delta code! */
typedef struct {
	/* Input to code generator */
	int prec;			/* Internal precision:- either 8 or 16 bits */ 
	int id;				/* Number of input dimensions */
	int od;				/* Number of output dimensions */
	imdi_pixrep irep;	/* High level input pixel representation */
	imdi_pixrep orep;	/* High level output pixel representation */
	int in_signed;		/* Bit flag per channel, NZ if treat as signed (runtime setup) */
	int out_signed;		/* Bit flag per channel, NZ if treat as signed (runtime setup) */
	pixlayout in;		/* Input pixel layout */
	pixlayout out;		/* Output pixel layout */
	imdi_options opt;	/* Direction and stride options */
	int itres;			/* Interpolation table resolution */
	int stres;			/* Simplex table resolution */

	/* Returned value */
	char kkeys[100];		/* Kernel keys */
	char kdesc[100];		/* At genspec time */
	char kname[100];		/* At generation time */
} genspec;

/* - - - - - - - - - - - - - - - - - - - - - - - */

/* Translate between high level and low level generation specification */
int set_genspec(genspec *gs, gendesc *gd, int comb, mach_arch *a);

/* - - - - - - - - - - - - - - - - - - - - - - - */
/* Convert the input and output pixel representation and the internal */
/* precision choice into a precision in bits */
/* (Used at code gen and code execution) */
#define COMPUTE_IPREC(OUT, INREP, IPREC, OUTREP)	\
{													\
	int iprec, oprec;								\
													\
	switch (INREP) {								\
		default:									\
		case pixint8:								\
		case planeint8:								\
			iprec = 8;								\
			break;									\
		case pixint16:								\
		case planeint16:							\
			iprec = 16;								\
			break;									\
	}												\
	switch (OUTREP) {								\
		default:									\
		case pixint8:								\
		case planeint8:								\
			oprec = 8;								\
			break;									\
		case pixint16:								\
		case planeint16:							\
			oprec = 16;								\
			break;									\
	}												\
	switch (IPREC) {								\
		default:									\
		case prec_min:								\
			OUT = iprec < oprec ? iprec : oprec;	\
			break;									\
		case prec_max:								\
			OUT = iprec > oprec ? iprec : oprec;	\
			break;									\
		case prec_in:								\
			OUT = iprec;							\
			break;									\
		case prec_out:								\
			OUT = oprec;							\
			break;									\
		case prec_p8:								\
			OUT = 8;								\
			break;									\
		case prec_p16:								\
			OUT = 16;								\
			break;									\
	}												\
}													\

/* - - - - - - - - - - - - - - - - - - - - - - - */

struct _tabspec;

/* Supported code generators: */

/* The 'C' code generator */
int gen_c_kernel(genspec *g, struct _tabspec *t, mach_arch *a,
                 FILE *fp, int index, genspec *og, struct _tabspec *ot);

/* asm, MMX, etc. generators declarations go here ! */

#endif /* IMDI_GEN_H */











