
#ifndef ICOMS_H

/* 
 * Argyll Color Correction System
 *
 * An abstracted instrument serial and USB communication class.
 *
 * Author: Graeme W. Gill
 * Date:   2006/4/20
 *
 * Copyright 1996 - 2006 Graeme W. Gill
 * All rights reserved.
 *
 * This material is licenced under the GNU GENERAL PUBLIC LICENCE :-
 * see the Licence.txt file for licencing details.
 * 
 * Derived from serio.h
 */

#ifdef ENABLE_USB
#include "usb.h"

#endif /* ENABLE_USB */

/* Flow control */
typedef enum {
	fc_nc,
	fc_none,
	fc_XonXOff,
	fc_Hardware
} flow_control;

/* baud rate available on all systems */
typedef enum {
	baud_nc      = 0,		/* Not Configured */
	baud_110     = 1,
	baud_300     = 2,
	baud_600     = 3,
	baud_1200    = 4,
	baud_2400    = 5,
	baud_4800    = 6,
	baud_9600    = 7,
	baud_14400   = 8,
	baud_19200   = 9,
	baud_38400   = 10,
	baud_57600   = 11,
	baud_115200  = 12
} baud_rate;

/* Possible parity */
typedef enum {
	parity_nc,
	parity_none,
	parity_odd,
	parity_even
} parity;

/* Possible stop bits */
typedef enum {
	stop_nc,
	stop_1,
	stop_2
} stop_bits;

/* Possible word length */
typedef enum {
	length_nc,
	length_5,
	length_6,
	length_7,
	length_8
} word_length;

/* Status bits/ return values */
#define ICOM_TO		0x1000		/* Timed out */
#define ICOM_USER 	0x2000		/* User abort */
#define ICOM_UTER 	0x4000		/* User terminated */

#define ICOM_USBR	0x0100		/* Unspecified USB read error */
#define ICOM_USBW	0x0200		/* Unspecified USB write error */
#define ICOM_SERR	0x0400		/* Unspecified Serial read error */
#define ICOM_SERW	0x0800		/* Unspecified Serial write error */

#define ICOM_XRE	0x0040		/* Xmit shift reg empty */
#define ICOM_XHE	0x0020		/* Xmit hold reg empty */
#define ICOM_BRK	0x0010		/* Break detected */
#define ICOM_FER	0x0008		/* Framing error */
#define ICOM_PER	0x0004		/* Parity error */
#define ICOM_OER	0x0002		/* Overun error */
#define ICOM_DRY	0x0001		/* Recv data ready */

/* Store information about a possible instrument communication path */
typedef struct {
	char *path;					/* Serial port path, or USB device description */
#ifdef ENABLE_USB
	struct usb_device *dev;		/* USB port, NULL if not USB */
	instType itype;				/* Type of instrument on the USB port */
#endif /* ENABLE_USB */
} icompath;

#if defined (NT)
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif

struct _icoms {
  /* Private: */
	icompath *ppath;			/* path to selected port */
	
	int port;					/* io port number: 1, 2, 3, 4 .. n */
	int is_usb;					/* Flag, NZ if this is a USB port */
	int is_open;				/* Flag, NZ if this port is open */

	/* Serial port parameters */
#if defined (MSDOS)
	unsigned short porta;		/* Hardware port address */
#endif
#if defined (NT)
	HANDLE phandle;				/* NT handle */
#endif
#if defined (UNIX) || defined(__APPLE__)
	int fd;						/* Unix file descriptor */
#endif
	flow_control fc;
	baud_rate	baud;
	parity		parity;
	stop_bits	stop_bits;
	word_length	word_length;

#ifdef ENABLE_USB
	/* USB port parameters */
	usb_dev_handle *usbh;
	int cnfg;				/* Configuration */
	int ifce;				/* Interface */
	int wr_ep, rd_ep;		/* End points to use */
	int rd_qa;				/* Read quanta size */
#endif /* ENABLE_USB */

	/* General parameters */
	int lerr;					/* Last error code */
	int tc;						/* Current termination character (-1 if not set) */
	int npaths;
	icompath **paths;			/* Paths if any */
	int debug;					/* Flag, nz to print instrument io info to stderr */
	int utc;					/* User termination character, 0 if none */

	/* Linked list to automate SIGKILL cleanup */
	struct _icoms *next;
	
  /* Public: */

	/* Return a list of instrument io communication paths */
	/* Return NULL on failure. */
	icompath ** (*get_paths)(		/* Return pointer to list of icompath */
		struct _icoms *p 			/* End is marked with NULL pointer, */
	);								/* Storage will be freed with icoms object */

	/* Return the instrument type if the port number is USB, instUnknown if not */
	instType (*is_usb_portno)(
		struct _icoms *p, 
		int 	 	 port);		/* Enumerated port number, 1..n */

	/* Select the serial communications port and characteristics */
	void (*set_ser_port)(
		struct _icoms *p, 
		int 	 	 port,		/* Enumerated port number, 1..n */
		flow_control fc,		/* Flow control */
		baud_rate	 baud,
		parity		 parity,
		stop_bits	 stop_bits,
		word_length	 word_length);

	/* Select the USB communications port and characteristics */
	void (*set_usb_port)(
		struct _icoms *p, 
		int            port,	/* Enumerated port number, 1..n */
	    int            config,	/* Configuration */
	    int            iface,	/* Interface */
	    int            wr_ep,	/* Write end point */
	    int            rd_ep,	/* Read end point */
	    int            rd_qa);	/* Read quanta size */

	/* Set a user termination character, 0 for none */
	void (*set_user_term)(struct _icoms *p, int ch);

	/* Write the characters in the buffer out */
	/* Data will be written up to the terminating nul. */
	/* Return relevant error status bits */
	int (*write)(
		struct _icoms *p,
		char *buf,
		double tout);		/* Timeout in seconds */

	/* Read characters into the buffer */
	/* The returned data will be terminated by a nul. */
	int (*read)(
		struct _icoms *p,
		char *buf,			/* Buffer to store characters read */
		int bsize,			/* Buffer size */
		char tc,			/* Terminating character */
		int ntc,			/* Number of terminating characters */
		double tout);		/* Timeout in seconds */

	/* write and read */
	int (*write_read)(
		struct _icoms *p,
		char *wbuf,			/* Write puffer */
		char *rbuf,			/* Read buffer */
		int bsize,			/* Buffer size */
		char tc,			/* Terminating characer */
		int ntc,			/* Number of terminating characters */
		double tout);		/* Timeout in seconds */

	/* Destroy ourselves */
	void (*del)(struct _icoms *p);

}; typedef struct _icoms icoms;

/* Constructor */
extern icoms *new_icoms(void);

/* - - - - - - - - */
/* Othe functions */

/* Create and return a list of available serial ports or USB instruments for this system. */
/* We look at the registry key "HKEY_LOCAL_MACHINE\HARDWARE\DEVICEMAP\SERIALCOMM" */
/* to determine serial ports, and use libusb to discover USB instruments. */
/* Return NULL if function fails */
icompath **icoms_get_paths(icoms *p); 

/* - - - - - - - - - - - - - - - - - - -- */
/* System dependent convenience functions */

/* wait for and then return the next character from the keyboard */
int next_con_char(void);

/* If here is one, return the next character from the keyboard, else return 0 */
int poll_con_char(void);

/* Empty the console of any pending characters */
void empty_con_chars(void);

#define ICOMS_H
#endif /* ICOMS_H */
